"""
This module provides a wrapper for interacting with an API.

The module contains the `ApiWrapper` class which offers convenient methods to access various
endpoints of the API using HTTP requests. It supports custom headers and proxies for
additional flexibility.
"""
import json
import logging
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry


logger = logging.getLogger()


class InvalidApiResponse(Exception):
    pass


class ApiRequestException(Exception):
    pass


class ApiWrapper:
    """
    Wrapper class for interacting with an API.

    This class provides convenient methods to access various endpoints of the
    API using HTTP requests. It supports custom headers and proxies for
    additional flexibility.

    Args:
        base_url (str): The base URL of the API.
        headers (dict, optional): Dictionary containing customer headers to be sent with each request.
        proxies (dict, optional): Dictionary containing proxy settings to be used for requests.
        debug (bool, optional): Setting this to true will print `urllib3` debug information.

    Attributes:
        base_url (str): The base URL of the API.
        session (requests.Session): Persistent HTTP session.
    """
    def __init__(self, base_url, headers=None, proxies=None, debug=False):
        self.base_url = base_url
        self.session = requests.Session()

        retries = Retry(total=3, backoff_factor=1, status_forcelist=[429, 500, 502, 503, 504])
        self.session.mount('http://', HTTPAdapter(max_retries=retries))
        self.session.mount('https://', HTTPAdapter(max_retries=retries))

        if headers:
            self.session.headers.update(headers)
        if proxies:
            self.session.proxies.update(proxies)

        if debug:
            if not logger.handlers:
                logging.basicConfig()

            logging.getLogger("urllib3").setLevel(logging.DEBUG)

    def __get(self, endpoint, params=None):
        url = self.base_url + endpoint

        try:
            response = self.session.get(url, params=params)
            response.raise_for_status()
        except requests.RequestException as e:
            raise ApiRequestException(f'Failed to execute GET request: {str(e)}')

        try:
            return response.json()
        except json.JSONDecodeError:
            raise InvalidApiResponse('Invalid JSON received')

    def __post(self, endpoint, data=None):
        url = self.base_url + endpoint

        try:
            response = self.session.post(url, data=data)
            response.raise_for_status()
        except requests.RequestException as e:
            raise ApiRequestException(f'Failed to execute POST request: {str(e)}')

        try:
            return response.json()
        except json.JSONDecodeError:
            raise InvalidApiResponse('Invalid JSON received')

    def __put(self, endpoint, data=None):
        url = self.base_url + endpoint

        try:
            response = self.session.put(url, data=data)
            response.raise_for_status()
        except requests.RequestException as e:
            raise ApiRequestException(f'Failed to execute PUT request: {str(e)}')

        try:
            return response.json()
        except json.JSONDecodeError:
            raise InvalidApiResponse('Invalid JSON received')

    def __patch(self, endpoint, data=None):
        url = self.base_url + endpoint

        try:
            response = self.session.put(url, data=data)
            response.raise_for_status()
        except requests.RequestException as e:
            raise ApiRequestException(f'Failed to execute PATCH request: {str(e)}')

        try:
            return response.json()
        except json.JSONDecodeError:
            raise InvalidApiResponse('Invalid JSON received')

    def __delete(self, endpoint):
        url = self.base_url + endpoint

        try:
            response = self.session.delete(url)
            response.raise_for_status()
        except requests.RequestException as e:
            raise ApiRequestException(f'Failed to execute DELETE request: {str(e)}')

        try:
            return response.json()
        except json.JSONDecodeError:
            raise InvalidApiResponse('Invalid JSON received')

    def get_users(self):
        endpoint = '/users'
        response = self.__get(endpoint)
        return response
